@@ -21,6 +21,10 @@ the shebang:
 ...to the top of your file, and executing it directly like a shell
 script. Note that this requires you to have zxpy installed globally.
 """
+
+from __future__ import annotations
+
+import argparse
 import ast
 import code
 import codecs
@@ -30,11 +34,16 @@ import shlex
 import subprocess
 import sys
 import traceback
-from typing import Generator, Tuple, Union, IO
+from typing import Any, Generator, IO, Optional
 
 UTF8Decoder = codecs.getincrementaldecoder("utf8")
 
 
+class ZxpyArgs(argparse.Namespace):
+    interactive: Optional[bool]
+    filename: str
+
+
 def cli() -> None:
     """
     Simple CLI interface.
@@ -47,22 +56,117 @@ def cli() -> None:
 
         zxpy
     """
-    # Remove zxpy executable from argv
-    del sys.argv[0]
+    parser = argparse.ArgumentParser()
+    parser.add_argument(
+        "-i",
+        "--interactive",
+        action="store_true",
+        help="Run in interactive mode",
+    )
+    parser.add_argument("filename", help="Name of file to run", nargs="?")
 
-    if len(sys.argv) == 0:
+    # Everything passed after a `--` is arguments to be used by the script itself.
+    script_args = ["/bin/sh"]
+    try:
+        separator_index = sys.argv.index("--")
+        script_args.extend(sys.argv[separator_index + 1 :])
+        # Remove everything after `--` so that argparse passes
+        sys.argv = sys.argv[:separator_index]
+    except ValueError:
+        # `--` not present in command, so no extra script args
+        pass
+
+    args = parser.parse_args(namespace=ZxpyArgs())
+
+    # Once arg parsing is done, replace argv with script args
+    sys.argv = script_args
+
+    if args.filename is None:
         setup_zxpy_repl()
         return
 
-    filename = sys.argv[0]
-    with open(filename) as file:
+    with open(args.filename) as file:
         module = ast.parse(file.read())
-        run_zxpy(filename, module)
+
+        globals_dict: dict[str, Any] = {}
+        try:
+            run_zxpy(args.filename, module, globals_dict)
+        except Exception:
+            # Only catch the exception in interactive mode
+            if not args.interactive:
+                raise
+
+            traceback.print_exc()
+
+        if args.interactive:
+            globals().update(globals_dict)
+            install()
+
+
+def is_inside_single_quotes(string: str, index: int) -> bool:
+    """Returns True if the given index is inside single quotes in a shell command."""
+    quote_index = string.find("'")
+    if quote_index == -1:
+        # No single quotes
+        return False
+
+    if index < quote_index:
+        # We're before the start of the single quotes
+        return False
+
+    double_quote_index = string.find('"')
+    if double_quote_index >= 0 and double_quote_index < quote_index:
+        next_double_quote = string.find('"', double_quote_index + 1)
+        if next_double_quote == -1:
+            # Double quote opened but never closed
+            return False
+
+        # Single quotes didn't start and we passed the index
+        if next_double_quote >= index:
+            return False
+
+        # Ignore all single quotes inside double quotes.
+        index -= next_double_quote + 1
+        rest = string[next_double_quote + 1 :]
+        return is_inside_single_quotes(rest, index)
+
+    next_quote = string.find("'", quote_index + 1)
+    if next_quote >= index:
+        # We're inside single quotes
+        return True
+
+    index -= next_quote + 1
+    rest = string[next_quote + 1 :]
+    return is_inside_single_quotes(rest, index)
 
 
 @contextlib.contextmanager
 def create_shell_process(command: str) -> Generator[IO[bytes], None, None]:
     """Creates a shell process, yielding its stdout to read data from."""
+    # shell argument support, i.e. $0, $1 etc.
+
+    dollar_indices = [index for index, char in enumerate(command) if char == "$"]
+    for dollar_index in reversed(dollar_indices):
+        if (
+            dollar_index >= 0
+            and dollar_index + 1 < len(command)
+            and command[dollar_index + 1].isdigit()
+            and not is_inside_single_quotes(command, dollar_index)
+        ):
+            end_index = dollar_index + 1
+            while end_index + 1 < len(command) and command[end_index + 1].isdigit():
+                end_index += 1
+
+            number = int(command[dollar_index + 1 : end_index + 1])
+
+            # Get argument number from sys.argv
+            if number < len(sys.argv):
+                replacement = sys.argv[number]
+            else:
+                replacement = ""
+
+            command = command[:dollar_index] + replacement + command[end_index + 1 :]
+
     process = subprocess.Popen(
         command,
         stdout=subprocess.PIPE,
@@ -89,15 +193,14 @@ def run_shell_print(command: str) -> None:
     """Version of `run_shell` that prints out the response instead of returning a string."""
     with create_shell_process(command) as stdout:
         decoder = UTF8Decoder()
-        with open(stdout.fileno(), 'rb', closefd=False) as buff:
-            # read1 should be present on BinaryIO but it's not. Typeshed bug
-            for text in iter(buff.read1, b""):  # type: ignore
+        with open(stdout.fileno(), "rb", closefd=False) as buff:
+            for text in iter(buff.read1, b""):
                 print(decoder.decode(text), end="")
 
             print(decoder.decode(b"", final=True), end="")
 
 
-def run_shell_alternate(command: str) -> Tuple[str, str, int]:
+def run_shell_alternate(command: str) -> tuple[str, str, int]:
     """Like run_shell but returns 3 values: stdout, stderr and return code"""
     process = subprocess.Popen(
         command,
@@ -118,23 +221,31 @@ def run_shell_alternate(command: str) -> Tuple[str, str, int]:
     )
 
 
-def run_zxpy(filename: str, module: ast.Module) -> None:
+def run_zxpy(
+    filename: str,
+    module: ast.Module,
+    globals_dict: dict[str, Any] | None = None,
+) -> None:
     """Runs zxpy on a given file"""
     patch_shell_commands(module)
     code = compile(module, filename, mode="exec")
-    exec(
-        code,
+
+    if globals_dict is None:
+        globals_dict = {}
+
+    globals_dict.update(
         {
             "__name__": "__main__",
             "$run_shell": run_shell,
             "$run_shell_alternate": run_shell_alternate,
             "$run_shell_print": run_shell_print,
             "$shlex_quote": shlex.quote,
-        },
+        }
     )
+    exec(code, globals_dict)
 
 
-def patch_shell_commands(module: Union[ast.Module, ast.Interactive]) -> None:
+def patch_shell_commands(module: ast.Module | ast.Interactive) -> None:
     """Patches the ast module to add zxpy functionality"""
     shell_runner = ShellRunner()
     shell_runner.visit(module)
@@ -149,12 +260,8 @@ def quote_fstring_args(fstring: ast.JoinedStr) -> None:
             if (
                 isinstance(node.format_spec, ast.JoinedStr)
                 and len(node.format_spec.values) == 1
-                and (
-                    isinstance(node.format_spec.values[0], ast.Str)
-                    and node.format_spec.values[0].s == "raw"
-                    or isinstance(node.format_spec.values[0], ast.Constant)
-                    and node.format_spec.values[0].value == "raw"
-                )
+                and isinstance(node.format_spec.values[0], ast.Constant)
+                and node.format_spec.values[0].value == "raw"
             ):
                 node.format_spec = None
                 continue
@@ -178,7 +285,7 @@ class ShellRunner(ast.NodeTransformer):
         if (
             isinstance(expr, ast.UnaryOp)
             and isinstance(expr.op, ast.Invert)
-            and isinstance(expr.operand, (ast.Str, ast.JoinedStr))
+            and isinstance(expr.operand, (ast.Constant, ast.JoinedStr))
         ):
             if isinstance(expr.operand, ast.JoinedStr):
                 quote_fstring_args(expr.operand)
@@ -235,6 +342,35 @@ def setup_zxpy_repl() -> None:
     print()
 
     install()
+    sys.exit()
+
+
+class ZxpyConsole(code.InteractiveConsole):
+    """Runs zxpy over"""
+
+    def runsource(
+        self,
+        source: str,
+        filename: str = "<console>",
+        symbol: str = "single",
+    ) -> bool:
+        # First, check if it could be incomplete input, return True if it is.
+        # This will allow it to keep taking input
+        with contextlib.suppress(SyntaxError, OverflowError):
+            if code.compile_command(source) is None:
+                return True
+
+        try:
+            ast_obj = ast.parse(source, filename, mode=symbol)
+            assert isinstance(ast_obj, ast.Interactive)
+            patch_shell_commands(ast_obj)
+            code_obj = compile(ast_obj, filename, mode=symbol)
+        except (ValueError, SyntaxError):
+            # Let the original implementation take care of incomplete input / errors
+            return super().runsource(source, filename, symbol)
+
+        self.runcode(code_obj)
+        return False
 
 
 def install() -> None:
@@ -253,17 +389,8 @@ def install() -> None:
     if len(frames) > 1:
         parent_frame = frames[1]
         parent_locals = parent_frame.frame.f_locals
-        locals().update(parent_locals)
-
-    # setup zxpy globals
-    globals().update(
-        {
-            "$run_shell": run_shell,
-            "$run_shell_alternate": run_shell_alternate,
-            "$run_shell_print": run_shell_print,
-            "$shlex_quote": shlex.quote,
-        }
-    )
+    else:
+        parent_locals = {}
 
     # For tab completion and arrow key support
     if sys.platform != "win32":
@@ -271,63 +398,15 @@ def install() -> None:
 
         readline.parse_and_bind("tab: complete")
 
-    command = ""
-    continued_command = False
-    while True:
-        try:
-            prompt = "... " if continued_command else ">>> "
-            new_input = input(prompt)
-        except KeyboardInterrupt:
-            print()
-            if continued_command:
-                continued_command = False
-                command = ""
-            continue
-        except EOFError:
-            print()
-            sys.exit(0)
+    zxpy_locals = {
+        **parent_locals,
+        "$run_shell": run_shell,
+        "$run_shell_alternate": run_shell_alternate,
+        "$run_shell_print": run_shell_print,
+        "$shlex_quote": shlex.quote,
+    }
 
-        # TODO: refactor the next 10 lines.
-        # probably move command = '...' stuff somewhere else
-        if continued_command:
-            command += "\n"
-        else:
-            command = ""
-
-        if new_input != "":
-            command += new_input
-        else:
-            continued_command = False
-
-        if continued_command:
-            continue
-
-        try:
-            ast_obj = ast.parse(command, "<input>", "single")
-        except SyntaxError:
-            try:
-                code_obj = code.compile_command(command)
-                if code_obj is None:
-                    continued_command = True
-                    continue
-
-            except BaseException:
-                traceback.print_exc()
-                continue
-
-        assert isinstance(ast_obj, ast.Interactive)
-        patch_shell_commands(ast_obj)
-
-        try:
-            code_obj = compile(ast_obj, "<input>", "single")
-            assert code_obj is not None
-            exec(code_obj)
-
-        except SystemExit as e:
-            sys.exit(e.code)
-
-        except BaseException:
-            traceback.print_exc()
+    ZxpyConsole(locals=zxpy_locals).interact(banner="", exitmsg="")
 
 
 if __name__ == "__main__":
